from abc import ABC, abstractmethod
from .components_manager import ComponentsManager
from .config_block import ConfigBlock
from .injector import Injector
import inspect

import time


class Platform(ABC):

    def __init__(self, manager, env_config, behaviour):
        self.manager = manager
        self.env_config = env_config
        self.behaviour = behaviour
        self.config = env_config['platform'] if 'platform' in env_config else ConfigBlock({})

        self.injector = Injector(self, self.manager.components_tree)
        self.plugins = [self.init_plugin(pl_tuple) for pl_tuple in env_config['plugins']]

    def init_plugin(self, plugin_tuple):

        if callable(plugin_tuple):
            return plugin_tuple

        plugin_cls = plugin_tuple[0] if type(plugin_tuple) is tuple else plugin_tuple
        plugin_conf = ConfigBlock(plugin_tuple[1] if type(plugin_tuple) is tuple else {})

        injector = Injector(self, self.manager.components_tree, plugin_conf)
        class_members = {t[0]: t[1] for t in inspect.getmembers(plugin_cls)}
        init_members = {t[0]: t[1] for t in inspect.getmembers(class_members['__init__'])}
        init_annotations = init_members['__annotations__'] if '__annotations__' in init_members else {}

        return plugin_cls(**injector.get_all(init_annotations))

    @staticmethod
    def create(config: ConfigBlock):

        from ..platforms.mjc.platform import PlatformMJC

        config = ConfigBlock(config)
        config.defaults({
            'defaults': {
                'environment': 'sim',
                'behaviour': {

                },
                'plugins': [

                ]
            },
            'environments': {
                'sim': {
                    'platform': {
                        'class': PlatformMJC
                    },
                    'components': {

                    }
                }
            }
        })

        world = config['world']
        behaviour = config['behaviour']
        env = config['defaults']['environment']
        env_config = config['environments'][env]
        env_config.defaults(config['defaults'] if 'defaults' in config else {})

        manager = ComponentsManager(
            world,
            env_config['components']
        )

        platform_class = env_config['platform']['class'] \
            if env_config.is_config_block('platform') else env_config['platform']
        return platform_class(manager, env_config, behaviour)

    @abstractmethod
    def step(self):
        pass

    def root(self):
        return self.manager.components_tree['instance']

    def init_components(self, interfaces):
        """
            Initializes components.
            Called after all the interfaces are instantiated (by the platform manager)

            Iterates through all components,
                - Gets the component class annotations,
                instantiates the component and stores at component['instance']

                - Gets and stores the component interface
        """

        def step_components(comp):
            # build the children, its dependencies, first.
            for child in comp['children']:
                step_components(child)

            c_data = comp['class'].__data__
            try:
                comp['injector'] = injector = Injector(self, comp)
                comp['instance'] = c = comp['class'](**{
                    n: injector.get(n, c_data['annotations'][n]) for n in c_data['annotations']
                })

                if comp['id'] in interfaces:
                    interface = c_data['interface_instance'] = interfaces[comp['id']]
                    methods = [i for i in dir(interface.__class__) if i[0:2] != '__']
                    [setattr(c, m, getattr(interface, m)) for m in methods if hasattr(c, m)]

            except TypeError as e:
                raise TypeError(str(e) + ', while initiating ' + str(comp['class']) + (
                    '/' + comp['name'] if comp['name'] is not None else ''))

        step_components(self.manager.components_tree)

    def run(self):

        if self.behaviour is not None:
            class_members = {t[0]: t[1] for t in inspect.getmembers(self.behaviour)}
            init_members = {t[0]: t[1] for t in inspect.getmembers(class_members['__init__'])}
            init_annotations = init_members['__annotations__'] if '__annotations__' in init_members else {}

            bh_config = self.env_config['behaviour']
            bh_injector = Injector(self, self.manager.components_tree, config=bh_config)
            bh = self.behaviour(**bh_injector.get_all(init_annotations))
            if hasattr(bh, 'on_start'):
                bh.on_start()

            if hasattr(bh, 'on_update'):
                while True:
                    if bh.on_update() == False:
                        break
                    self.wait_once()
        else:
            self.wait_forever()

    def wait(self, fn=None, cb=None):
        while True:
            self.step()
            if self.plugins:
                [(plugin() if callable(plugin) else plugin.step()) for plugin in self.plugins]
            if (fn is None) \
                    or (isinstance(fn, list) and not any([not e() for e in fn])) \
                    or (callable(fn) and fn()):
                break
            if cb is not None:
                cb()

    def wait_seconds(self, seconds, cb=None):
        start_seconds = time.time()
        self.wait(lambda: time.time() - start_seconds > seconds, cb)

    def wait_forever(self, cb=None):
        self.wait(lambda: False, cb)

    def wait_once(self, cb=None):
        self.wait(None, cb)



import cv2

import time
import numpy as np
from yarok.core.config_block import ConfigBlock

import mujoco_viewer

def normalize(img):
    img1 = img - np.min(img)
    mx = np.max(img1)
    if mx < 1e-4:
        return img1
    return img1 / mx


class ViewerMJC:
    """
        The MuJoCo viewer wrapper.

        The step method can be called to update the viewer.

    """

    def __init__(self, platform, config: ConfigBlock):
        self.model = platform.model
        self.data = platform.data
        self.config = ConfigBlock(config)
        self.config.defaults({
            'refresh_rate': 1. / 33,
            'width': None,
            'height': None,
            'hide_menus': True
        })


        self.next_probe_ts = time.time() + self.config['refresh_rate']

        self.viewer = mujoco_viewer.MujocoViewer(self.model,
                                                 self.data,
                                                 width=self.config['width'],
                                                 height=self.config['height'],
                                                 hide_menus=self.config['hide_menus'],
                                                 )

    def step(self):
        if time.time() > self.next_probe_ts:
            self.next_probe_ts = time.time() + self.config['refresh_rate']
            self.viewer.render()



# from mujoco import MjRenderContextOffscreen, MjRenderContextWindow, MjRenderContext

import numpy as np
import mujoco


class InterfaceMJC:
    """
        This Interface serves as a simplified MuJoCo api,
        it handles lower level naming/referencing and offers an object/component oriented interface.

        It is instantiated and passed to the constructor of each corresponding component MuJoCo's interface.
    """

    def __init__(self, name, platform):
        global offscreen_id
        self.component_name = name
        self.platform = platform
        self.sensor_name2adr = {}
        self.actuator_name2adr = {}
        self.cam_name2adr = {}
        self.sensor_name2idx = {}
        self.actuator_name2idx = {}
        self.cam_name2idx = {}
        self.actuators = []
        self.sensors = []
        self.cams = []
        self.component_name = ''
        self.contexts = {}
        self.scn = mujoco.MjvScene(self.platform.model, maxgeom=10000)
        self.vopt = mujoco.MjvOption()
        self.pert = mujoco.MjvPerturb()

    def on_init(self):
        self.contexts = {
            cam_name: mujoco.MjrContext(self.platform.model, mujoco.mjtFontScale.mjFONTSCALE_150.value)
            for cam_name in self.cam_name2adr
        }

    # https://github.com/htung0101/table_dome/blob/master/table_dome_calib/utils.py#L160
    def depthimg2Meters(self, depth):
        extent = self.platform.model.stat.extent
        near = self.platform.model.vis.map.znear * extent
        far = self.platform.model.vis.map.zfar * extent
        image = near / (1 - depth * (1 - near / far))
        return image

    def set_ctrl(self, k, v):
        actuator_name = self.actuators[k] if type(k) == int else k
        self.platform.data.ctrl[self.actuator_name2idx[actuator_name]] = v

    def sensordata(self):
        return [self.platform.data.sensordata[self.sensor_name2idx[sn]] for sn in self.sensors]

    def read_camera(self, camera_name, shape=(480, 640), depth=False, rgb=True):
        """
            mj_(offscreen)context.render() api:
            https://github.com/openai/mujoco-py/blob/master/mujoco_py/mjrendercontext.pyx#L131

            mj_(offscreen)context.read_pixels() api:
            https://github.com/openai/mujoco-py/blob/master/mujoco_py/mjrendercontext.pyx#L176

            and for ref, mjSim render()
            https://github.com/openai/mujoco-py/blob/master/mujoco_py/mjsim.pyx#L131
        """
        context = self.contexts[camera_name]
        camera_id = self.cam_name2idx[camera_name]

        self.depth_arr = np.zeros(shape, dtype=np.float32)
        self.rgb = np.zeros(shape + (3,), dtype=np.uint8)

        viewport = mujoco.MjrRect(0, 0, shape[1], shape[0])

        cam = mujoco.MjvCamera()
        cam.type = 2
        cam.fixedcamid = camera_id

        mujoco.mjv_updateScene(
            self.platform.model,
            self.platform.data,
            self.vopt,
            self.pert,
            cam,
            mujoco.mjtCatBit.mjCAT_ALL.value,
            self.scn)

        mujoco.mjv_defaultCamera(cam)
        mujoco.mjr_render(viewport, self.scn, context)
        mujoco.mjr_readPixels(self.rgb, self.depth_arr, viewport, context)

        self.rgb = np.fliplr(self.rgb)
        self.depth_arr = np.fliplr(self.depth_arr)

        if not depth:
            return self.rgb
        d = self.depth_arr if depth == 'raw' else self.depthimg2Meters(self.depth_arr)

        if not rgb:
            return d

        return self.rgb, d



from yarok import component, interface, ConfigBlock
from yarok.platforms.mjc import InterfaceMJC

import numpy as np
import cv2
import os
import open3d as o3d

import math

from time import time

from .sim_model.model import SimulationModel
from .sim_model.utils.camera import circle_mask


@interface(
    defaults={
        'frame_size': (320, 240),
    }
)
class GelTipInterfaceMJC:

    def __init__(self, interface_mjc: InterfaceMJC, config: ConfigBlock):
        self.interface = interface_mjc
        self.frame_size = config['frame_size']
        self.mask = circle_mask(self.frame_size)
        bkg_zeros = np.zeros(self.frame_size[::-1] + (3,), dtype=np.float32)

        __location__ = os.path.realpath(
            os.path.join(os.getcwd(), os.path.dirname(__file__)))
        assets_path = os.path.join(__location__, './sim_model/assets/')

        elastic_deformation = True
        cloud, geodesic_light_fields = SimulationModel.load_assets(
            assets_path,
            (160, 120),
            self.frame_size,
            'geodesic',
            3)

        light_coeffs = [
            {'color': [196, 94, 255], 'id': 0.5, 'is': 0.1},  # red # [108, 82, 255]
            {'color': [154, 144, 255], 'id': 0.5, 'is': 0.1},  # green # [255, 130, 115]
            {'color': [104, 175, 255], 'id': 0.5, 'is': 0.1},  # blue  # [120, 255, 153]
        ]

        light_fiels = [{'field': geodesic_light_fields[l], **light_coeffs[l]} for l in range(3)]

        self.model = SimulationModel(**{
            'ia': 0.8,
            'light_sources': light_fiels,
            'background_depth': cv2.resize(np.load(assets_path + 'bkg.npy'), self.frame_size),
            'cloud_map': cloud,
            'background_img': bkg_zeros,  # bkg_rgb if use_bkg_rgb else
            'elastomer_thickness': 0.004,
            'min_depth': 0.026,
            'texture_sigma': 0.000005,
            'elastic_deformation': elastic_deformation
        })
        self.depth = np.zeros(self.frame_size, np.float32)

        self.last_update = 0

    def read(self):
        t = time()
        if self.last_update > t - 1.0:
            return self.tactile
        self.last_update = t
        shape = self.frame_size[1], self.frame_size[0]
        self.depth = self.interface.read_camera('camera', shape, depth=True, rgb=False)
        self.tactile = self.model.generate(self.depth).astype(np.uint8)
        return self.tactile

    def read_depth(self):
        return self.depth


class GelTipInterfaceHW:

    def __init__(self, config: ConfigBlock):
        self.cap = cv2.VideoCapture(config['cam_id'])
        if not self.cap.isOpened():
            raise Exception('GelTip cam ' + str(config['cam_id']) + ' not found')

        self.fake_depth = np.zeros((640, 480), np.float32)

    def read(self):
        [self.cap.read() for _ in range(10)]  # skip frames in the buffer.
        ret, frame = self.cap.read()
        return frame

    def read_depth(self):
        return self.fake_depth


@component(
    tag="geltip",
    defaults={
        'interface_mjc': GelTipInterfaceMJC,
        'interface_hw': GelTipInterfaceHW,
        'probe': lambda c: {'camera': c.read()},
        # template configs,
        'label_color': '255 0 0'
    },
    # language=xml
    template="""
        <mujoco>
            <asset>
                <material name="glass_material" rgba="1 1 1 0.1"/>
                <material name="white_elastomer" rgba="1 1 1 1"/>
                <material name="black_plastic" rgba=".3 .3 .3 1"/>
                
                <material name="label_color" rgba="${label_color} 1.0"/>
        
                <mesh name="geltip_shell" file="meshes/shell_open.stl" scale="0.001 0.001 0.001"/>
                <mesh name="geltip_sleeve" file="meshes/sleeve_open.stl" scale="0.001 0.001 0.001"/>
                <mesh name="geltip_mount" file="meshes/mount.stl" scale="0.001 0.001 0.001"/>
                
                <!-- the glass -->
                <mesh name="geltip_glass" file="meshes/glass_long.stl" scale="0.00099 0.00099 0.00099"/>
                
                <!-- the outter elastomer, for visual purposes -->
                <mesh name="geltip_elastomer" file="meshes/elastomer_long.stl" scale="0.0011 0.0011 0.0011"/>  
                
                <!-- inverted mesh, for limiting the depth map-->
                <!-- changing this mesh changes the depth maps -->
                <mesh name="geltip_elastomer_inv" file="meshes/elastomer_long_inv.stl" scale="0.00105 0.00105 0.00105"/>
        
            </asset>
            <worldbody>
                <body name="geltip">
                    <geom type="sphere" 
                        density="0.1"
                        material="label_color" 
                        size="0.005" 
                        pos="0.0 0.012 -0.025"/>
                    <geom density="0.1" type="mesh" mesh="geltip_shell" material="black_plastic"/>
                    <geom density="0.1" type="mesh" mesh="geltip_sleeve" material="black_plastic"/>
                    <geom density="0.1" type="mesh" mesh="geltip_mount" material="black_plastic"/>
                    <camera name="camera" pos="0 0 0.01" zaxis="0 0 -1" fovy="70"/>
                    <body>
                    
                       <!-- mesh, to serve as the glass and detect collisions -->
                             <!-- solimp="1.0 1.2 0.001 0.5 2" 
                              solref="0.02 1"-->
                       <geom density="0.1" type="mesh" 
                              mesh="geltip_glass" 
                              pos="0.0 0.0 -0.003"
                              friction="1 0.05 0.01" 
                              solimp="1.1 1.2 0.001 0.5 2" solref="0.02 1"
                              material="glass_material" /> 
                              
                       <!-- inverted, mesh, for limiting the depth-map -->
                       <!-- changing this geom/mesh changes the depth maps -->
                       <!-- 32 for contype and conaffinity disable collisions -->     
                       <geom  type="mesh" 
                              mesh="geltip_elastomer_inv" 
                              contype="32"  
                              conaffinity="32" 
                              pos="0.0 0.0 -0.005"
                              material="white_elastomer" />
                       
                       <!-- white elastomer, for visual purposes -->
                       <geom density="0.1" 
                              type="mesh" 
                              mesh="geltip_elastomer" 
                              friction="1 0.05 0.01" 
                              contype="32" 
                              conaffinity="32" 
                              pos="0.0 0.0 -0.007"
                              material="white_elastomer"/>
                    </body>
        
                </body>
            </worldbody>
        </mujoco>
    """
)
class GelTip:

    def __init__(self):
        """
            Geltip driver as proposed in
            https://danfergo.github.io/geltip-simulation/

            The frame method gets the depth map from the simula
        """
        pass

    def read(self):
        pass

    def read_depth(self):
        pass



import os

__location__ = os.path.realpath(
    os.path.join(os.getcwd(), os.path.dirname(__file__)))

import subprocess

try:
    from .ikfastpy import ikfastpy
except (ModuleNotFoundError, ImportError) as e:
    print('To enable UR5e pose control, Andy Zeng ikfastpy is going to be cloned and installed.')
    print('Insert [Y] to continue... else abort.')
    if input('') != 'Y':
        abort(-1)

    try:
        import Cython
    except ModuleNotFoundError as e:
        print('We need to install Cython. Using pip or pip3?')
        builder = ('' + input('')).lower().strip()
        if builder != 'pip' and builder != 'pip3':
            exit(-1)
        os.system(f'{builder} install Cython')

    print('Cloning https://github.com/andyzeng/ikfastpy...')
    os.system(f'cd {__location__} && \
                  rm -rf ikfastpy && \
                  git clone https://github.com/andyzeng/ikfastpy.git')

    print('Build with python or python3?')
    builder = ('' + input('')).lower().strip()
    if builder != 'python' and builder != 'python3':
        exit(-1)

    os.system(f'cd {__location__}/ikfastpy && \
                  {builder} setup.py build_ext --inplace')

    print('Done. Finished installing ikfastpy.')

    # import it again, it should be fine now.
    from .ikfastpy import ikfastpy

from yarok import Platform, PlatformMJC, PlatformHW, component, interface, ConfigBlock, Injector

from yarok.platforms.mjc import InterfaceMJC

from math import pi, sin, cos
import time
import numpy as np
from scipy.spatial.transform import Rotation as R


def sae(q1, q2):
    return sum([abs(q1[i] - q2[i]) for i in range(6)])


@interface(
    # max-speed https://store.clearpathrobotics.com/products/ur5e
    defaults={
        'speed': pi / 24
    }
)
class UR5eInterfaceMJC:
    def __init__(self, interface: InterfaceMJC, config: ConfigBlock):
        self.interface = interface
        self.speed = config['speed']
        initial_q = [
            0,
            0,
            0,
            0,
            0,
            0
        ]

        self.ws = [
            [- pi, pi],  # shoulder pan
            [- pi, 0],  # shoulder lift,
            [- 2 * pi, 2 * pi],  # elbow
            [-2 * pi, 2 * pi],  # wrist 1
            [- 2 * pi, 2 * pi],  # wrist 2
            [- 2 * pi, 2 * pi]  # wrist 3
        ]
        self.stopped_steps = 0

        self.q = initial_q
        self.start_q = initial_q
        self.target_q = initial_q
        self.start_t = time.time()
        self.delta_t = 0

        self.ur5_kin = ikfastpy.PyKinematics()
        self.n_joints = self.ur5_kin.getDOF()

        # transformation_matrix = np.array([[0.04071115, -0.99870914, 0.03037599, 0.3020009],
        #                                   [-0.99874455, -0.04156303, -0.02796067, 0.12648243],
        #                                   [0.0291871, -0.02919955, -0.99914742, 0.53451169]])
        # print(self.get_pose_vectors(transformation_matrix))

        # ee_pose = HTrans(np.array([self.q]).T, [0])  # self.ur5_kin.forward(self.q)
        # ee_pose = np.asarray(ee_pose)
        # ee_pose = np.asarray(ee_pose).reshape(3, 4)  # 3x4 rigid transformation matrix
        # print('INITIAL transformation matrix ')
        # print(ee_pose)
        # rotation_matrix = ee_pose[0:3, 0:3]
        # translation_matrix = ee_pose[0:3, 3].T
        # xyz = R.from_matrix(rotation_matrix).as_euler('xyz')
        # print('Initial translation', translation_matrix, 'Initial rotation (xyz): ', xyz)

    def set_ws(self, ws):
        self.ws = ws

    def set_speed(self, speed):
        self.speed = speed

    def get_transformation_matrix(self, q):
        ee_pose = self.ur5_kin.forward(q)
        ee_pose = np.asarray(ee_pose).reshape(3, 4)
        return ee_pose

    def get_pose_vectors(self, transformation_matrix):
        # print(ee_pose)
        rotation_matrix = transformation_matrix[0:3, 0:3]
        xyz = transformation_matrix[0:3, 3].T
        xyz_angles = R.from_matrix(rotation_matrix).as_euler('xyz')
        # print('----')
        # print(current_xyz)
        # print(current_xyz_angles)
        # print('----')
        # print(R.from_euler('xyz', current_xyz_angles).as_matrix())
        return xyz, xyz_angles

    def compute_transformation_matrix(self, xyz, xyz_angles):
        rotation_matrix = R.from_euler('xyz', xyz_angles).as_matrix()
        return np.concatenate([rotation_matrix, np.array([xyz]).T], axis=1)

    def compute_ik(self, transformation_matrix):

        joint_configs = self.ur5_kin.inverse(transformation_matrix.reshape(-1).tolist())
        n_solutions = int(len(joint_configs) / self.n_joints)
        joint_configs = np.asarray(joint_configs).reshape(n_solutions, self.n_joints)

        if n_solutions == 0:
            print('[Error] no solutions found for this pose')
            return None

        valid_config = lambda config: not any([not (self.ws[i][0] < config[i] < self.ws[i][1]) for i in range(6)])
        valid_joint_configs = [config for config in joint_configs if valid_config(config)]

        if len(valid_joint_configs) == 0:
            print('Configs outside the workspace:')
            print(joint_configs)
            print('[Error] no solutions within the workspace')
            return None

        closest_config_index = np.argmin([sae(self.q, q) for q in valid_joint_configs])
        q = valid_joint_configs[closest_config_index]
        return q

    def ik(self, xyz=None, xyz_angles=None):
        c_transformation_matrix = self.get_transformation_matrix(self.q)
        current_xyz, current_xyz_angles = self.get_pose_vectors(c_transformation_matrix)
        q_xyz = xyz if xyz is not None else current_xyz
        q_xyz_angles = xyz_angles if xyz_angles is not None else current_xyz_angles
        transformation_matrix = self.compute_transformation_matrix(q_xyz, q_xyz_angles)
        return self.compute_ik(transformation_matrix)

    def move_xyz_delta(self, xyz=None, xyz_angles=None):
        c_transformation_matrix = self.get_transformation_matrix(self.q)
        current_xyz, current_xyz_angles = self.get_pose_vectors(c_transformation_matrix)
        new_xyz = current_xyz if xyz is None else np.add(current_xyz, xyz)
        new_xyz_angles = current_xyz_angles if xyz_angles is None else np.add(current_xyz_angles, xyz_angles)
        q = self.ik(new_xyz, new_xyz_angles)

        if q is not None:
            return self.move_q(q)

    def move_xyz(self, xyz=None, xyz_angles=None):
        q = self.ik(xyz, xyz_angles)

        if q is not None:
            return self.move_q(q)

        # ee_pose = HTrans(np.array([self.q]).T, [0])  # self.ur5_kin.forward(self.q)
        # ee_pose = np.asarray(ee_pose)
        # current_rotation_matrix = ee_pose[0:3, 0:3]
        # current_translation_vector = ee_pose[0:3, 3].T
        # current_xyz_angles = R.from_matrix(current_rotation_matrix).as_euler('xyz')
        # print('Current translation', current_translation_vector, 'rotation (xyz): ', current_xyz_angles)
        #
        # # rotation_matrix = euler_angles([0, 1.6, 0])
        # # print(rotation_matrix)
        # # rotation_matrix = R.from_euler('xyz', [1.6/2, 0, 1.6]).as_matrix()
        # translation_vector = np.array([current_translation_vector if xyz is None else xyz]).T
        # rotation_matrix = current_rotation_matrix if xyz_angles is None else R.from_euler('xyz', xyz_angles).as_matrix()
        # transformation_matrix = np.concatenate([rotation_matrix, translation_vector], axis=1)
        # # print(translation_vector.T.shape, rotation_matrix.shape)
        # # print('transformation matrix', transformation_matrix)
        # # rot_matrix = np.concatenate([, ], axis=1)
        # # rot_matrix = rot_matrix[..., np.newaxis]
        #
        # # transformation_matrix = np.array([[0.04071115, -0.99870914, 0.03037599, 0.3020009],
        # #                     [-0.99874455, -0.04156303, -0.02796067, 0.12648243],
        # #                     [0.0291871, -0.02919955, -0.99914742, 0.53451169]])
        # transformation_matrix_h = np.concatenate([transformation_matrix, np.array([[0, 0, 0, 1]])], axis=0)
        # joint_configs = invKine(transformation_matrix_h)
        # # joint_configs = self.ur5_kin.inverse(transformation_matrix.reshape(-1).tolist())
        # n_solutions = len(joint_configs)
        # # print('>>', joint_configs)
        # # print('--------------')
        # # print('>>', np.array(joint_configs[0])[0])
        # # n_solutions = int(len(joint_configs) / self.n_joints)
        # # joint_configs = np.asarray(joint_configs).reshape(n_solutions, self.n_joints)
        #
        # joint_configs = np.asarray(joint_configs)
        #
        # # print(n_solutions)
        # # # for joint_config in joint_configs:
        # # #     print(joint_config)
        # # print(joint_configs)
        # # print('-----')
        # # print(valid_joint_configs)
        # # print('-----')
        #
        # #
        # # print(n_solutions, closest_config_index, q)
        #
        # # q = [-3.1, -1.6, 1.6, -1.6, -1.6, 0.]
        # self.move_q(q)

    def move_q(self, q):
        self.start_q = self.target_q
        self.target_q = q
        self.start_t = time.time()
        self.delta_t = max([abs(self.target_q[i] - self.start_q[i]) / self.speed for i in range(6)])
        return lambda: self.is_at(q)

    def step(self):
        self.last_q = self.q
        self.q = self.interface.sensordata()
        now = time.time()
        # print(self.q[1])
        if self.start_q is not None:
            elapsed = (now - self.start_t)
            progress = min(1, elapsed / self.delta_t) if self.delta_t > 0 else 1
            target = [(1 - progress) * self.start_q[i] + progress * self.target_q[i] for i in range(6)]
            # print(target)
        else:
            target = self.target_q

        [self.interface.set_ctrl(a, target[a]) for a in range(len(self.interface.actuators))]

        if sae(self.last_q, self.q) < 1e-6:
            self.stopped_steps += 1
        else:
            self.stopped_steps = 0

    def at(self):
        return self.q

    def at_xyz(self):
        transformation_matrix = self.get_transformation_matrix(self.q)
        return self.get_pose_vectors(transformation_matrix)

        # return [q / self.gear for q in self.q]

    def is_at(self, q):
        # s = sum([abs(self.q[i] - q[i]) for i in range(6)])
        # print(s, self.q)
        # print(self.q)
        # print(sum([abs(self.q[i] - q[i]) for i in range(6)]))
        # print(sum([abs(self.q[i] - q[i]) for i in range(6)]))
        # print('-> ', sae(q, self.q))
        err = sae(q, self.q)
        return err < 0.006 or (err < 0.025 and self.stopped_steps >= 30)
        # def similar_q(q1, q2):
        #     print(sum([abs(q1[i] - q2[i]) for i in range(len(q1))]))
        # return sum([abs(q1[i] - q2[i]) for i in range(len(q1))]) < 0.05
        #
        # at = self.at()
        # if self.last_query_position is not None:
        #     if similar_q(q, self.last_query_position):
        #         self.at_target_hit_counter += 1
        #
        #         if self.at_target_hit_counter > 30:
        #             self.last_query_position = None
        #             self.at_target_hit_counter = 0
        #             return True
        #
        #     else:
        #         self.last_query_position = None
        #         self.at_target_hit_counter = max(self.at_target_hit_counter, 0)
        #
        # self.last_query_position = at
        #
        # return False


@component(
    tag='ur5e',
    defaults={
        'interface_mjc': UR5eInterfaceMJC
    },
    template_path='ur5e.xml'
)
class UR5e:

    def __init__(self):
        pass

    def move_xyz(self, xyz=None, xyz_angles=None):
        pass

    def move_xyz_delta(self, xyz=None, xyz_angles=None):
        pass

    def move_q(self, args):
        pass

    def at(self):
        pass

    def ik(self, xyz=None, xyz_angles=None):
        pass

    def at_xyz(self):
        pass

    def is_at(self):
        pass

    def set_ws(self, ws):
        pass

    def set_speed(self, speed):
        pass


class Behav:

    def __init__(self, pl: Platform, injector: Injector):
        self.pl = pl
        self.arm: UR5e = injector.get('ur5e')

    def on_update(self):
        q = [pi / 2, -pi / 2, pi / 2 - pi / 4, 0, pi / 2, pi / 2]
        self.arm.move_q(q)
        self.pl.wait(lambda: self.arm.is_at(q))
        self.pl.wait_seconds(5)
        # self.gripper.move(0)

        q = [pi / 2, -pi / 2, pi / 2, 0, pi / 2, pi / 2]
        self.arm.move_q(q)
        # self.pl.wait(lambda: self.arm.is_at(q))
        self.pl.wait_seconds(5)

        # self.gripper.move(0.78)

        # self.pl.wait_seconds(10)
        self.pl.wait_seconds(5)

        q = [pi / 2, -pi / 2, pi / 2 - pi / 10, 0, pi / 2, pi / 2]
        self.arm.move_q(q)
        self.pl.wait_seconds(5)
        # self.pl.wait(lambda: self.arm.is_at(q))

        q = [pi / 2, -pi / 2 + pi / 3, pi / 2 - pi / 10 + pi / 5, 0, pi / 2, pi / 2]
        self.arm.move_q(q)
        self.pl.wait_seconds(5)
        # self.pl.wait(lambda: self.arm.is_at(q))

        # self.gripper.move(0.5)


def test():
    @component(
        components=[
            UR5e
        ],
        # language=xml
        template="""
                <mujoco>
                    <compiler angle="radian" meshdir="assets" autolimits="true" />

                    <worldbody>
                           <ur5e name='ur5e'/>
                    </worldbody>
                </mujoco>
            """
    )
    class UR5eTestWorld:
        pass

    conf = {
        'world': UR5eTestWorld,
        'behaviour': Behav,
        'defaults': {
            'environment': 'sim',
            'components': {
                '/gripper': {
                    'left_tip': False,
                    'right_tip': False,
                }
            },
            'plugins': [
                # (Cv2Inspector, {})
            ]
        },
        'environments': {
            'sim': {
                'platform': {
                    'class': PlatformMJC,
                    'width': 1000,
                    'height': 800,
                }
            }
        },
    }
    Platform.create(conf).run()


if __name__ == '__main__':
    test()



"""
    Gomes' Gelsight driver

    This takes in a depth map and combines it with a background image (taken from a real
    gelsight sensor) and combines them both with mathematical techniques to produce
    a gelsight output.

    The code is a DIGIT-modified version of what can be found here and is mostly unchanged:
        https://github.com/danfergo/gelsight_simulation
"""


import cv2
import numpy                    as np
import scipy.ndimage.filters    as fi


def show_normalized_img(name, img):
    draw = img.copy()
    draw -= np.min(draw)
    draw = draw / np.max(draw)
    cv2.imshow(name, draw)
    return draw


def gkern2(kernlen=21, nsig=3):
    """Returns a 2D Gaussian kernel array."""

    # create nxn zeros
    inp = np.zeros((kernlen, kernlen))
    # set element at the middle to one, a dirac delta
    inp[kernlen // 2, kernlen // 2] = 1
    # gaussian-smooth the dirac, resulting in a gaussian filter mask
    return fi.gaussian_filter(inp, nsig)


def gaus_noise(image, sigma):
    row, col = image.shape
    mean = 0
    gauss = np.random.normal(mean, sigma, (row, col))
    gauss = gauss.reshape(row, col)
    noisy = image + gauss
    return noisy


def derivative(mat, direction):
    assert (direction == 'x' or direction == 'y'), "The derivative direction must be 'x' or 'y'"
    kernel = None
    if direction == 'x':
        kernel = [[-1.0, 0.0, 1.0]]
    elif direction == 'y':
        kernel = [[-1.0], [0.0], [1.0]]
    kernel = np.array(kernel, dtype=np.float64)
    return cv2.filter2D(mat, -1, kernel) / 2.0


def tangent(mat):
    dx = derivative(mat, 'x')
    dy = derivative(mat, 'y')
    img_shape = np.shape(mat)
    _1 = np.repeat([1.0], img_shape[0] * img_shape[1]).reshape(img_shape).astype(dx.dtype)
    unormalized = cv2.merge((-dx, -dy, _1))
    norms = np.linalg.norm(unormalized, axis=2)
    return (unormalized / np.repeat(norms[:, :, np.newaxis], 3, axis=2))


def solid_color_img(color, size):
    image = np.zeros(size + (3,), np.float64)
    image[:] = color
    return image


def add_overlay(rgb, alpha, color):
    s = np.shape(alpha)

    opacity3 = np.repeat(alpha, 3).reshape((s[0], s[1], 3))  # * 10.0

    overlay = solid_color_img(color, s)

    foreground = opacity3 * overlay
    background = (1.0 - opacity3) * rgb.astype(np.float64)
    res = background + foreground

    res[res > 255.0] = 255.0
    res[res < 0.0] = 0.0
    res = res.astype(np.uint8)

    return res


# DIGIT Simulation
class PhongRender:
    def __init__(self, **config):
        self.light_sources = config['light_sources']
        self.background = config['background_img']
        self.px2m_ratio = config['px2m_ratio']
        self.elastomer_thickness = config['elastomer_thickness']
        self.min_depth = config['min_depth']

        self.default_ks = 0.15
        self.default_kd = 0.5
        self.default_alpha = 5

        self.ka = config['ka'] or 0.8

        self.texture_sigma = config['texture_sigma'] or 0.00001
        self.t = config['t'] if 't' in config else 3
        self.sigma = config['sigma'] if 'sigma' in config else 7
        self.kernel_size = config['sigma'] if 'sigma' in config else 21

        self.max_depth = self.min_depth + self.elastomer_thickness


    def protrusion_map(self, original, not_in_touch):
        protrusion_map = np.copy(original)
        protrusion_map[not_in_touch >= self.max_depth] = self.max_depth
        return protrusion_map


    def segments(self, depth_map):
        
        not_in_touch = np.copy(depth_map)
        not_in_touch[not_in_touch < self.max_depth] = 0.0
        not_in_touch[not_in_touch >= self.max_depth] = 1.0
        in_touch = 1 - not_in_touch

        return not_in_touch, in_touch


    def internal_shadow(self, elastomer_depth):
        elastomer_depth_inv = self.max_depth - elastomer_depth
        elastomer_depth_inv = np.interp(elastomer_depth_inv, (0, self.elastomer_thickness), (0.0, 1.0))
        return elastomer_depth_inv


    def apply_elastic_deformation_v1(self, protrusion_depth, not_in_touch, in_touch):
        kernel = gkern2(15, 7)
        deformation = self.max_depth - protrusion_depth

        for i in range(5):
            deformation = cv2.filter2D(deformation, -1, kernel)

        return 30 * -deformation * not_in_touch + (protrusion_depth * in_touch)


    def apply_elastic_deformation(self, protrusion_depth, not_in_touch, in_touch):
        protrusion_depth = - (protrusion_depth - self.max_depth)

        kernel = gkern2(self.kernel_size, self.sigma)
        deformation = protrusion_depth
        deformation2 = protrusion_depth
        kernel2 = gkern2(52, 9)

        for i in range(self.t):
            deformation_ = cv2.filter2D(deformation, -1, kernel)
            r = np.max(protrusion_depth) / np.max(deformation_) if np.max(deformation_) > 0 else 1
            deformation = np.maximum(r * deformation_, protrusion_depth)

            deformation2_ = cv2.filter2D(deformation2, -1, kernel2)
            r = np.max(protrusion_depth) / np.max(deformation2_) if np.max(deformation2_) > 0 else 1
            deformation2 = np.maximum(r * deformation2_, protrusion_depth)

        # deformation_v1 = self.apply_elastic_deformation_v1(protrusion_depth, not_in_touch, in_touch)

        # for i in range(self.t):
        #     deformation_ = cv2.filter2D(deformation2, -1, kernel)
        #     r = np.max(protrusion_depth) / np.max(deformation_) if np.max(deformation_) > 0 else 1
        #     deformation2 = np.maximum(r * deformation_, protrusion_depth)


        deformation_x = 2.0 * deformation - deformation2
      

        return self.max_depth - deformation_x


    # use phong illuminations to 
    def phong_illumination(self, T, source_dir, kd, ks, alpha):
        dot = np.dot(T, np.array(source_dir)).astype(np.float64)
        difuse_l = dot * kd
        difuse_l[difuse_l < 0] = 0.0

        dot3 = np.repeat(dot[:, :, np.newaxis], 3, axis=2)

        R = 2.0 * dot3 * T - source_dir
        V = [0.0, 0.0, 1.0]

        spec_l = np.power(np.dot(R, V), alpha) * ks

        return difuse_l + spec_l


    # main worker class
    # -- call and pass depth map in order to create a simulated digit image
    def generate(self, obj_depth, return_depth=False):
        
        not_in_touch, in_touch      = self.segments(obj_depth)
        protrusion_depth            = self.protrusion_map(obj_depth, not_in_touch)
        elastomer_depth             = self.apply_elastic_deformation(protrusion_depth, not_in_touch, in_touch)

        textured_elastomer_depth    = gaus_noise(elastomer_depth, self.texture_sigma)

        out = self.ka * self.background
        out = add_overlay(out, self.internal_shadow(protrusion_depth), (0.0, 0.0, 0.0))

        T = tangent(textured_elastomer_depth / self.px2m_ratio)

        for light in self.light_sources:
            ks      = light['ks'] if 'ks' in light else self.default_ks
            kd      = light['kd'] if 'kd' in light else self.default_kd
            alpha   = light['alpha'] if 'alpha' in light else self.default_alpha
            out     = add_overlay(out, self.phong_illumination(T, light['position'], kd, ks, alpha), light['color'])

        kernel = gkern2(3, 1)
        out = cv2.filter2D(out, -1, kernel)


        if return_depth:
            return out, elastomer_depth
        return out


import importlib
import os
from os import path
import os
import argparse

from .__init__ import run, __VERSION__

import importlib.util
import sys
from enum import Enum

from .cli import generate_template


# adapted from https://stackoverflow.com/questions/19053707/converting-snake-case-to-lower-camel-case-lowercamelcase
def to_camel_case(snake_str):
    components = snake_str.split('_')
    # We capitalize the first letter of each component except the first one
    # with the 'title' method and join them together.
    return ''.join(x.title() for x in components)


def load(p):
    """
    Dynamically loads a module with name "name" and retrieves
    the class with the same name declared in that file.
    :param p:
    :param class_name:
    :return: loaded class.
    """

    module_name = p[p.rfind('.') + 1:]

    try:
        spec = importlib.util.spec_from_file_location(module_name, os.getcwd() + '/' + p.replace('.', '/') + ".py")
        module = importlib.util.module_from_spec(spec)
        sys.modules[module_name] = module
        spec.loader.exec_module(module)
    except FileNotFoundError:
        module = importlib.import_module('.components.' + p, __package__)
    return getattr(module, to_camel_case(module_name))


def main():
    """
        yarok cli entry point.
        Parses the cli arguments and calls the corresponding python methods/apis, e.g.,

        yarok run World
            ->  yarok.run(config)
    """

    parser = argparse.ArgumentParser()
    parser.add_argument('command', choices=['g', 'run'])
    parser.add_argument('arg0', type=str)

    parser.add_argument('-V', '--version', action='store_true', default=False,
                        help='display version information and exit')

    args = parser.parse_args()

    if args.command == 'g':
        generate_template(args.arg0)
    elif args.version:
        print('Version ' + __VERSION__)
    else:
        print("We don't have a help menu here. Check our pipy or github.")

if __name__ == '__main__':
    main()
